/**********************************************************************
 mps.c :   Make a PostScript file out of a plain ASCII text file

 Version 2.1.0   June 1999

 PostScript is a registered trademark of Adobe Systems Incorporated.

 Copyright (C) 1999, Richard Urena.  All rights reserved.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.   You can find the GNU
 General Public License in the file COPYING in the source directory,
 or at http://www.gnu.org

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#include "defaults.h"
#include "options.h"
#include "globals.h"

/* general globals */

static int AbsolutelyNothingDone = 1;


/*   pagination-related globals and macros */

#define  TopOfPage   ((Paper->Height)-Margin.Top.Measure-(Font->Ascend))
static int PageIsEmpty = 1;
static float Y;
static int PageCounter = 0;

/*	input and output line buffers */

static char *LineBuffer;
static char *OutputLine;

/* line-folding globals */

static int MaxOutputChars;
static float ContinueX;
static int CleanBreakLimit;

/*  macro to output a string in postscript */

char *PSfmt = "(%s) %3.1f %3.1f S\n";
#define	PSLineOut(w,x,y)	printf(PSfmt,w,x,y)

/*   ANSI function prototypes   */

int mps_environmentoptions(char *);
int mps_commandlineoptions(int, char **);
int CheckArguments(int, char **);
void PSWriteTrailer(void);
void PSWriteProlog(void);
void WriteFile(FILE *, char *);
int GetLine(char *, FILE *, char *, int *, int *);
void ExpandTabs(char *, int *);
void PSWriteLine(char *, int);
void PSEject(void);
void PSBeginPage(void);
void PSStartBlankPage(void);
int AtBottom(void);
void trim(char *);
void PSPageNumber(void);

/* ------------------------------------------------------------------
   m a i n 
   ------------------------------------------------------------------ */
int main(int argc, char **argv)
{
	FILE *input;
	int i;

	if(!mps_environmentoptions(argv[0]) || !mps_commandlineoptions(argc,argv))
		exit(ExitCode);

	if(!ValidMargins(argv[0]))
		exit(ExitCode);

	LineBuffer = malloc(MPS__MAXLINESIZE);
	OutputLine = malloc(MPS__MAXLINESIZE);
	if((LineBuffer==NULL)||(OutputLine==NULL)) {
		mps_syserror(MPS__ERRNOMEM,ENOMEM,argv[0]);
		exit(ExitCode);
	}

	/* initalize various globals */
	Y = TopOfPage;
	MaxOutputChars = (int)((Paper->Width-Margin.Left.Measure
						- Margin.Right.Measure) / Font->CharWidth);
	ContinueX = Margin.Left.Measure - 2*(Font->CharWidth);
	CleanBreakLimit = MPS__CLEANBREAKLIMIT;
	if(CleanBreakLimit>(MaxOutputChars/2))		/* sanity check */
		CleanBreakLimit = 0;
	if((TabStop<1)||(TabStop>16))				/* ditto */
		TabStop = 8;

	/* let's go */
	PSWriteProlog();
	switch(i=CheckArguments(argc,argv)) {

		case 0:			 /* no input file specified, reads from stdin */
			WriteFile(stdin,"(stdin)");
			break;

		default:		/* otherwise, processes all input files, separating
						   them with page ejects */
			for( ; i<argc; i++) {
				if((input=fopen(argv[i],"r"))!=NULL) {
					WriteFile(input,argv[i]);
					fclose(input);
				} else
					mps_syserror(MPS__ERRNOFILE,errno,argv[i]);
			}
			break;
	}
	if(AbsolutelyNothingDone)
		PSEject();		/* otherwise, the printer will hang */
	PSWriteTrailer();

	exit(ExitCode);
}

/* ------------------------------------------------------------------
   G e t L i n e
   ------------------------------------------------------------------ */
int GetLine(char *line, FILE *input, char *filename, int *iscont,
		int *truncated)
{
	char c;
	int i, j, k, Last, Next;
	static int Embedded_FF = 0;
	static int Last_FF = 0;
	static int LineContinues = 0;
	static int LastLineBreak = 0;
#ifdef MPS__QUITONCTRLD
	static int ControlD = 0;
#endif
	int EndNow;


	(*iscont) = LineContinues;	 /* set by previous call */

	/*
	 * what we do at this point depends on the outcome of
	 * the previous call to this function
	 */

	EndNow = 0;
#ifdef MPS__QUITONCTRLD
	if(ControlD) {
		EndNow = 1;
	} else
#endif
	if(Embedded_FF) { 
		PSEject();
		PSBeginPage();
		i = Last_FF+1;
	} else if(LineContinues) {
		i = LastLineBreak;
	} else {
		errno = 0;
		if((fgets(LineBuffer,MPS__MAXLINESIZE,input)==NULL)||(errno!=0)) {
			EndNow = 1;
			if(errno!=0)
				mps_syserror(MPS__ERRFGETS,errno,filename);
		} else {
			trim(LineBuffer);
			ExpandTabs(LineBuffer,truncated);
			i = 0;
		}
	}

	if(EndNow) {
#ifdef MPS__QUITONCTRLD
		ControlD = 0;
#endif
		Embedded_FF = 0;
		LineContinues = 0;
		return(0);
	}

	/*
	 * at this point we have an input line (or portion thereof)
	 * on the input buffer, and the tabs have been expanded.
	 *
	 * "i" points to the beginning of the line (offset into
	 * the input buffer).
	 *
	 * we need to see if what remains of this line is still 
	 * too long, and if so, break up the line accordingly.
	 */

	LineContinues = (strlen(&(LineBuffer[i])) > MaxOutputChars);

	if(LineContinues) {
		/*
		 * search backwards from the
		 * break for a space to fold at, subject to
		 * CleanBreakLimit.  i.e., we're trying to fold
		 * the line somewhat elegantly.
		 */
		for(j=(MaxOutputChars-1),k=0; j>=0; j--,k++)
			if(((c=LineBuffer[i+j])==' ')||(k>=CleanBreakLimit))
				break;
		if(FoldLong&&(j>=0)&&(k<CleanBreakLimit)) {
			Next = (i+j+1);
			for( ; (j>0)&&((c=LineBuffer[i+j])==' '); j--)
				/* nothing */ ;
			Last = (j+i);
		} else {
			/*
			 * elegance didn't work, or is not wanted, so just chop the
			 * thing and be done with it
			 */
			Last = (i+MaxOutputChars-1);
			Next = Last+1;
		}
	}

	/* transfer the line (or portion thereof) to be printed */
	for(j=0; ((c=LineBuffer[i])!=MPS__PAGEEJECT)
#ifdef MPS__QUITONCTRLD
									&&(c!=MPS__CTRLD)
#endif
									&&(c!='\0'); i++,j++) {
		line[j] = c;
		if(LineContinues&&(i>=Last)) {
			++j;
			break;
		}
	}
	line[j] = '\0';

	/*
	 * finally, keep a record of the current state, for use
	 * by the next call to the function
	 */
#ifdef MPS__QUITONCTRLD
	ControlD = (c==MPS__CTRLD);
#endif
	Embedded_FF = (c==MPS__PAGEEJECT);
	if(Embedded_FF)
		Last_FF = i;

	/*
	 * if we are truncating instead of folding, turn off some
	 * warnings
	 */
	if(!FoldLong) {
		LineContinues = 0;
		truncated = 0;
	}

	if(LineContinues)
		LastLineBreak = Next;

	return(1);
}

/* ------------------------------------------------------------------
	 t r i m
	 get rid of the trailing newline (left there by fgets)
	 and any trailing whitespace on the input line
   ------------------------------------------------------------------ */
void trim(char *s)
{
	int i;

	for(i=(strlen(s)-1); (i>=0)&&((s[i]==' ')||(s[i]=='\t')||(s[i]=='\n'));
							i--)
		/* nothing */ ;
	s[i+1] = '\0';
}

/* ------------------------------------------------------------------
	 E x p a n d T a b s
	 tabs are expanded into corresponding spaces.

	 this only works with monospaced fonts
   ------------------------------------------------------------------ */
void ExpandTabs(char *src, int *truncated)
{
	char target[MPS__MAXLINESIZE];
	int i, j, c, prev, k, togo;

	/*
	 * if a tab is encountered, make sure the character
	 * immediately preceding it is a space.  if not,
	 * insert a space prior to expanding the tab
	 *
	 * we also need to check against buffer overflow.
	 * this is done only once per iteration, but that's
	 * good enough, given the generous buffer size..
	 */

	for(i=0,j=0,prev=0; (c=src[i])!='\0'; i++,prev=c) {
		if((j+(TabStop+1))>(MPS__MAXLINESIZE-1)) {
			(*truncated) = 1;
			break;
		}
		if(c!='\t')
			target[j++] = c;
		else {
			if(prev!=' ')
				target[j++] = ' ';
			togo = (TabStop-(j%TabStop))%TabStop;
			for(k=0; k<togo; k++)
				target[j++] = ' ';
		}
	}
	target[j] = '\0';
	strcpy(src,target);
}

/* ------------------------------------------------------------------
	 P S W r i t e L i n e
   ------------------------------------------------------------------ */
void PSWriteLine(char *line, int iscont)
{
	int i, j, c, blank, firstout;
	float X;

	if(AtBottom()) {
		PSEject();
		PSBeginPage();
	}

	if(iscont)
		PSLineOut(MPS__CONTINUECHAR,ContinueX,Y);

	for(blank=1,c=0,i=0,j=0;
		(j<(MPS__MAXLINESIZE-1))&&((c=(unsigned char)line[i])!='\0'); i++) {

		/* control chars. not previously dealt with are thrown out */
		if(c<' ')
			continue;

		/* don't include leading withespace as such */
		if(blank&&(c!=' ')) {
			firstout = i;
			blank = 0;
		}

		if(c>127) {
			/* per Adobe's recommendation, make the output 7-bit only */
			if((j+4)<=(MPS__MAXLINESIZE-1)) {
				sprintf(&(OutputLine[j]),"\\%03o",(unsigned char)c);
				j += 4;
			} else
				break;
		} else {
			/* things that need quoting inside a PostScript string */
			switch(c) {
				case '(':
				case ')':
				case '\\':
					OutputLine[j++] = '\\';
					/* FALLTHRU */
				default:
					OutputLine[j++] = c;
			}
		}
	}
	OutputLine[j] = '\0';

	if(!blank) {
		X = Margin.Left.Measure + firstout*(Font->CharWidth);
		PSLineOut(&(OutputLine[firstout]),X,Y);
	}

	if(iscont || (!blank))
		AbsolutelyNothingDone = 0;

	PageIsEmpty = 0;
	Y = (Y - (Font->LineSkip));


}

/* ------------------------------------------------------------------
   P S B e g i n P a g e
   ------------------------------------------------------------------ */
void PSBeginPage(void)
{
	printf("\n%%%%Page: %d\n",PageNo);
	PageCounter++;
	Y = TopOfPage;
}

/* ------------------------------------------------------------------
   P S E j e c t
   ------------------------------------------------------------------ */
void PSEject(void)
{
	if(PageNumbers)
		PSPageNumber();
	++PageNo;
	printf("P\n");
	Y = TopOfPage;
	AbsolutelyNothingDone = 0;
}

/* ------------------------------------------------------------------
   P S P a g e N u m b e r
   page numbers are centered across the bottom of the page
   ------------------------------------------------------------------ */
void PSPageNumber(void)
{
	float x, y;
	char temp[64];

	snprintf(temp,64,"%d",PageNo);
	temp[63] = '\0';
	x = (Paper->Width - (strlen(temp)*Font->CharWidth)) / 2.0;
	y = MPS__BOTTOMGUTTER;
	PSLineOut(temp,x,y);
}

/* ------------------------------------------------------------------
	 A t B o t t o m
   ------------------------------------------------------------------ */
int AtBottom(void)
{
	return((Y-(Font->LineSkip))<Margin.Bottom.Measure);
}

/* ------------------------------------------------------------------
   P S S t a r t B l a n k P a g e
   ------------------------------------------------------------------ */
void PSStartBlankPage(void)
{
	PSBeginPage();
	PSLineOut(" ",Margin.Left.Measure,Y);
	PageIsEmpty = 0;
}

/* ------------------------------------------------------------------
	 W r i t e F i l e
   ------------------------------------------------------------------ */
void WriteFile(FILE *input, char *filename)
{
	char line[MPS__MAXLINESIZE];
	int iscontinuation, truncated;

	truncated = 0;
	if(AssumeDuplex&&(PageCounter!=0)&&(PageCounter%2)) {
		PSStartBlankPage();
		PSEject();
	}
	printf("\n%% ------ begin file: %s ------ %%\n",filename);
	PSBeginPage();
	while(GetLine(line,input,filename,&iscontinuation,&truncated))
		PSWriteLine(line,iscontinuation);
	if(!PageIsEmpty)
		PSEject();
	if(truncated)
		mps_error(MPS__LINESTRUNCATED,filename);
}

/* ------------------------------------------------------------------
   m p s _ e n v i r o n m e n t o p t i o n s
   ------------------------------------------------------------------ */

/* beware.  misuse of strtok() can be hazardous to your health */

#define		EnvVarLength		256
#define		EnvDelimiter		" 	"	/* a space and a tab */

int mps_environmentoptions(char *Progname)
{
	char *temp, opts[EnvVarLength];

	if((temp=getenv(EnvironmentVar))!=NULL) {
		strncpy(opts,temp,(EnvVarLength-1));
		opts[EnvVarLength-1] = '\0';
		for(temp=strtok(opts,EnvDelimiter); temp!=NULL ;
								temp=strtok((char *)0,EnvDelimiter)) {
			if(!ParseOption(temp,Progname))
				return(0);
		}
	}
	return(1);
}

#undef EnvVarLength
#undef EnvDelimiter

/* ------------------------------------------------------------------
   m p s _ c o m m a n d l i n e o p t i o n s
   ------------------------------------------------------------------ */
int mps_commandlineoptions(int argc, char *argv[])
{
	int i;

	for(i=1; (i<argc)&&(argv[i][0]==MPS__OPTIONPREFIX); i++) {
		if(!ParseOption(argv[i],argv[0]))
			return(0);
	}
	return(1);
}

/* ------------------------------------------------------------------
	 C h e c k A r g u m e n t s
	 finds the first argument that is not an option
   ------------------------------------------------------------------ */
int CheckArguments(int argcounter, char **argument)
{
	int i;

	if(argcounter<=1)
		return(0);
	else {
		for(i=1; i<argcounter; i++) {
			if(argument[i][0]!=MPS__OPTIONPREFIX)
				return(i);
		}
	}
	return(0);
}

/* ------------------------------------------------------------------
   P S W r i t e P r o l o g
   ------------------------------------------------------------------ */
void PSWriteProlog(void)
{
	time_t now;

	now = time((time_t *)0);

	/* comments */

	printf("%%!PS-Adobe-3.0\n");
	printf("%%%%Creator: %s\n%%%%CreationDate: %s\n",VersionAsString(),
			ctime(&now));
	printf("%%%%DocumentFonts: %s\n",Font->Face);
	printf("%%%%LanguageLevel: 1\n%%%%Orientation: Portrait\n");
	printf("%%%%Pages: (atend)\n%%%%PageOrder: Ascend\n");
	printf("%%%%EndComments\n\n");

	/* prolog */

	printf("%%BeginProlog\n");
	printf("/S {moveto show} bind def\n");
	printf("/P {showpage} bind def\n");
	printf("%%EndProlog\n");

	/* setup */

	printf("\n%%%%BeginSetup\n");
	printf("%% Change the encoding vector to Latin-1 (ISO-8859-1)\n");
	printf("/%s findfont\n",Font->Face);
	printf("dup length dict begin\n");
	printf("	{1 index /FID ne {def} {pop pop} ifelse} forall\n");
	printf("	/Encoding ISOLatin1Encoding def\n");
	printf("	currentdict\n");
	printf("end\n");
	printf("/%s-ISOLatin1 exch definefont pop\n",Font->Face);
	printf("%% Scale to the desired size\n");
	printf("/%s-ISOLatin1 findfont %1.0f scalefont setfont\n",Font->Face,
		Font->PointSize);
	printf("%%%%EndSetup\n");
}

/* ------------------------------------------------------------------
   P S W r i t e T r a i l e r
   ------------------------------------------------------------------ */
void PSWriteTrailer(void)
{
	printf("\n%%%%Trailer\n%%%%Pages: %d\n%%%%EOF\n",PageCounter);
}

